import os
import pkg_resources
import json
import pyspark
import pyspark.sql.functions as F

from gluejobutils.s3 import read_json_from_s3
from gluejobutils.utils import read_json


def translate_metadata_type_to_type(column_type, target_type="glue"):
    """
    Uses json stored in package to convert generic meta data types to glue or spark data types.
    """
    io = pkg_resources.resource_stream(__name__, "data/data_type_conversion.json")
    lookup = json.load(io)

    try:
        return lookup[column_type][target_type]
    except:
        raise KeyError(
            "You attempted to lookup column type {}, but this cannot be found in data_type_conversion.json".format(
                column_type
            )
        )


def create_spark_schema_from_metadata(metadata, exclude_cols=[], non_nullable_cols=[]):
    """
    Creates a spark schema from a meta data dictionary.
    Excluding any column names that are in the exclude_cols list (default is an empty list).
    To specify any columns that should not be nullable list their names in non_nullable_cols (default is an empty list). 
    """
    columns = [c for c in metadata["columns"] if c["name"] not in exclude_cols]

    custom_schema = pyspark.sql.types.StructType()

    for c in columns:
        this_name = c["name"]
        nullable = this_name not in non_nullable_cols
        this_type = getattr(
            pyspark.sql.types, translate_metadata_type_to_type(c["type"], "spark")
        )
        this_field = pyspark.sql.types.StructField(this_name, this_type(), nullable)
        custom_schema.add(this_field)

    return custom_schema


def create_spark_schema_from_metadata_file(
    filepath, exclude_cols=[], non_nullable_cols=[]
):
    """
    Creates a spark schema from a json file that is a meta data dictionary. If filepath starts with s3:// the 
    function assumes it is an S3 file otherwise it tries to read the file from the local directory.
    """
    if "s3://" in filepath:
        metadata = read_json_from_s3(filepath)
    else:
        metadata = read_json(filepath)
    return create_spark_schema_from_metadata(
        metadata, exclude_cols=exclude_cols, non_nullable_cols=non_nullable_cols
    )


def align_df_to_meta(df, meta, exclude_columns=[], null_missing_cols=False):

    """
    Casts the columns in dataframe provided to the meta data dictionary provided
    """
    meta_cols = [m for m in meta["columns"] if m["name"] not in exclude_columns]

    df_cols = df.columns
    for m in meta_cols:
        this_type = getattr(
            pyspark.sql.types, translate_metadata_type_to_type(m["type"], "spark")
        )
        if m["name"] in df_cols:
            df = df.withColumn(m["name"], df[m["name"]].cast(this_type()))
        elif null_missing_cols:
            df = df.withColumn(m["name"], F.lit(None).cast(this_type()))
        else:
            raise ValueError(
                "ETL_ERROR: Column name in meta ({}) not in dataframe. Set null_missing_cols to True if you wish to null missing cols. Columns in dataframe {}".format(
                    m["name"], ", ".join(df_cols)
                )
            )

    df = df.select([x["name"] for x in meta_cols])

    return df



#   Module for some useful dataframe transforms
from itertools import chain
import pyspark.sql.functions as F


def apply_overwrite_dict_to_df(df, lookup_col, overwrite_dict):
    """
    df : A spark dataframe
    lookup_col : The column name that should be used to apply fixes (e.g. col1)
    overwrite_dict : should be a dictionary where each key is the value of lookup_col you want to fix. (e.g. {'a' : {'col2' : 2}} to fix the value in col2 when col1 is equal to 'a')
    """
    # Split overwrite_dict into a dictionary of single key value pairs dictionaries
    skvp = {}
    for k in overwrite_dict:
        for kk in overwrite_dict[k]:
            if kk not in skvp:
                skvp[kk] = {}

            skvp[kk][k] = overwrite_dict[k][kk]

    # for each col that is going to be overwritten apply the single key value pairs
    for k in skvp:
        mapping_expr = F.create_map([F.lit(x) for x in chain(*skvp[k].items())])
        df = df.withColumn(
            k,
            F.when(mapping_expr.getItem(df[lookup_col]).isNull(), df[k]).otherwise(
                mapping_expr.getItem(df[lookup_col])
            ),
        )

    return df



import json
import boto3
import botocore

from gluejobutils.utils import add_slash, remove_slash

try:
    # python 2.7
    from StringIO import StringIO
except:
    # python 3.6
    from io import StringIO

s3_resource = boto3.resource("s3")


def bucket_key_to_s3_path(bucket, key):
    """
    Takes an S3 bucket and key combination and returns the full S3 path to that location.
    """
    return "s3://{}/{}".format(bucket, key)


def s3_path_to_bucket_key(s3_path):
    """
    Splits out s3 file path to bucket key combination
    """
    s3_path = s3_path.replace("s3://", "")
    bucket, key = s3_path.split("/", 1)
    return bucket, key


def read_json_from_s3(s3_path, encoding="utf-8"):
    """
    read a json file from an s3 path
    """
    bucket, key = s3_path_to_bucket_key(s3_path)
    obj = s3_resource.Object(bucket, key)
    text = obj.get()["Body"].read().decode(encoding)
    return json.loads(text)


def write_json_to_s3(data, s3_path, indent=4, separators=(",", ": ")):
    """
    Saves your data to a json file (in memory and then sends it to the s3 path provided)
    """
    bucket, key = s3_path_to_bucket_key(s3_path)
    log_file = StringIO()
    json.dump(data, log_file, indent=indent, separators=separators)
    log_obj = s3_resource.Object(bucket, key)
    log_upload_resp = log_obj.put(Body=log_file.getvalue())
    return log_upload_resp


def get_filepaths_from_s3_folder(
    s3_folder_path, extension=None, exclude_zero_byte_files=True
):
    """
    Get a list of filepaths from a bucket. If extension is set to a string then only return files with that extension otherwise if set to None (default) all filepaths are returned.
    """
    if extension is None:
        extension = ""
    elif extension[0] != ".":
        extension = "." + extension

    s3_folder_path = add_slash(s3_folder_path)
    bucket, key = s3_path_to_bucket_key(s3_folder_path)

    s3b = s3_resource.Bucket(bucket)
    obs = s3b.objects.filter(Prefix=key)
    if exclude_zero_byte_files:
        ob_keys = [o.key for o in obs if o.key.endswith(extension) and o.size != 0]
    else:
        ob_keys = [o.key for o in obs if o.key.endswith(extension)]
    paths = sorted([bucket_key_to_s3_path(bucket, o) for o in ob_keys])

    return paths


def copy_s3_folder_contents_to_new_folder(old_s3_folder_path, new_s3_folder_path):
    """
    Copies complete folder structure within old_s3_folder_path to the new_s3_folder_path 
    """
    old_s3_folder_path = add_slash(old_s3_folder_path)
    new_s3_folder_path = add_slash(new_s3_folder_path)

    all_old_filepaths = get_filepaths_from_s3_folder(old_s3_folder_path)
    for ofp in all_old_filepaths:
        nfp = ofp.replace(old_s3_folder_path, new_s3_folder_path)
        copy_s3_object(ofp, nfp)


def delete_s3_object(s3_path):
    """
    Deletes the file at the s3_path given.
    """
    b, o = s3_path_to_bucket_key(s3_path)
    s3_resource.Object(b, o).delete()


def delete_s3_folder_contents(s3_folder_path):
    """
    Deletes all files within the s3_folder_path given given.
    """
    s3_folder_path = add_slash(s3_folder_path)
    all_filepaths = get_filepaths_from_s3_folder(
        s3_folder_path, exclude_zero_byte_files=False
    )
    for f in all_filepaths:
        delete_s3_object(f)


def copy_s3_object(old_s3_path, new_s3_path):
    """
    Copies a file in S3 from one location to another.
    """
    new_bucket, new_key = s3_path_to_bucket_key(new_s3_path)
    if "s3://" in old_s3_path:
        old_s3_path = old_s3_path.replace("s3://", "")
    s3_resource.Object(new_bucket, new_key).copy_from(CopySource=old_s3_path)


# https://stackoverflow.com/questions/33842944/check-if-a-key-exists-in-a-bucket-in-s3-using-boto3
def check_for_s3_file(s3_path):
    """
    Returns True/False depending if a file exists in the S3 path provided.
    exists = check_for_s3_file(s3://my-bucket/file.txt)
    """
    bucket, key = s3_path_to_bucket_key(s3_path)
    try:
        s3_resource.Object(bucket, key).load()
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        else:
            raise
    else:
        # The object does exist.
        return True


def folder_contains_only_files_with_extension(s3_folder_path, extension):
    """
    Checks if a folder contains any data (ingores zero-byte files) can also filter by extension.
    # Check if if folder only contains parquet files
    exists = folder_contains_only_files_with_extension("s3://my-bucket/data/", '.parquet').
    # Or check if only csv files exists in data
    exists = folder_contains_only_files_with_extension("s3://my-bucket/data/", '.csv')
    """
    if extension[0] != ".":
        extension = "." + extension
    all_p = get_filepaths_from_s3_folder(s3_folder_path)
    all_p_extension = [a for a in all_p if a.endswith(extension)]

    return len(all_p_extension) == len(all_p) and len(all_p_extension) != 0



# -*- coding: utf-8 -*-
import csv
import json
import os
import boto3
import botocore

from gluejobutils.s3 import (
    s3_path_to_bucket_key,
    folder_contains_only_files_with_extension,
)
from gluejobutils.utils import (
    remove_slash,
    add_slash,
    list_to_sql_select,
    standard_datetime_format,
    standard_datetime_format_python,
)

# No pyspark module if running on standard python docker so skip if need be
try:
    from pyspark.sql.types import *
    import pyspark.sql.functions as F
except ImportError:
    pass

from itertools import chain

static_record_end_datetime = "2999-01-01 00:00:00"


def set_record_start_datetime(
    df, datetime_string, datetime_format=standard_datetime_format, col_prefix=""
):
    """
    Set a start datetime to a df as the record_start_datetime. datetime_string should be a string representing datetime (default format is 'yyyy-MM-dd HH:mm:ss').
    For non standard format to datetime_string use datetime_format to put it in standard format.
    Note that the datetime_format specified should be a standard Java datetime format and not the python datetime format. 
    """
    record_start_datetime = col_prefix + "record_start_datetime"

    df = df.withColumn(
        record_start_datetime, F.to_timestamp(F.lit(datetime_string), datetime_format)
    )

    return df


def set_record_end_datetime(df, datetime_string=None, col_prefix=""):
    """
    Sets the column record_end_datetime the speficied datetime_string. Default value is to the datetime 2999-01-01 00:00:00.
    Returns a df.
    """
    record_end_datetime = col_prefix + "record_end_datetime"
    if datetime_string is None:
        datetime_string = static_record_end_datetime

    df = df.withColumn(
        record_end_datetime,
        F.to_timestamp(F.lit(datetime_string), standard_datetime_format),
    )
    return df


def init_record_datetimes(
    df, datetime_string, datetime_format=standard_datetime_format, col_prefix=""
):
    """
    Initialises record_start_datetime and record_end_datetime.
    Calls set_record_start_datetime and then set_record_end_datetime on df.
    Returns a df.
    """
    df = set_record_start_datetime(
        df=df,
        datetime_string=datetime_string,
        datetime_format=datetime_format,
        col_prefix=col_prefix,
    )
    df = set_record_end_datetime(df, col_prefix=col_prefix)
    return df


def update_record_end_datetime(
    df, partition_by, order_by, col_prefix="", filter_redundant_records=True
):
    """
    Takes dataframe and sets all record_end_datetime by ordering the data by specified order clause and partition variable.
    partition_by and order_by should be a string. If partitioning or ordering by multiple columns provide them as a single comma seperated string e.g. 'col1, col2'.
    record_end_datetime is set to the next records record_start_datetime. If the record_end_datetime is last row of that partition then it is set to 
    the default static_record_end_datetime (i.e. 2999-01-01 00:00:00).
    If filter_redundant_records is True (default) records in df with start_date == end_date will be filtered out.
    """
    start_col = col_prefix + "record_start_datetime"
    end_col = col_prefix + "record_end_datetime"

    df = df.withColumn(
        end_col,
        F.expr(
            "lead({}) OVER (PARTITION BY {} ORDER BY {})".format(
                start_col, partition_by, order_by
            )
        ),
    )

    # Set nulls to default record end date value
    df = df.withColumn(
        end_col,
        F.expr(
            "CASE WHEN {0} IS NULL THEN to_timestamp('{1}', '{2}') ELSE {0} END".format(
                end_col, static_record_end_datetime, standard_datetime_format
            )
        ),
    )

    # Filter out redundent records
    if filter_redundant_records:
        df = df.filter("{} <> {}".format(start_col, end_col))

    return df



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['gluejobutils']

package_data = \
{'': ['*'], 'gluejobutils': ['.ipynb_checkpoints/*', 'data/*']}

setup_kwargs = {
    'name': 'gluejobutils',
    'version': '2.0.0',
    'description': 'Python 2.7 utils for glue jobs',
    'long_description': '# gluejobutils\nPython 2.7 utility functions to include with AWS glue jobs\n',
    'author': 'Robin Linacre',
    'author_email': 'robinlinacre@hotmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
}


setup(**setup_kwargs)



import json

standard_date_format = "yyyy-MM-dd"
standard_date_format_python = "%Y-%m-%d"

standard_datetime_format = "yyyy-MM-dd HH:mm:ss"
standard_datetime_format_python = "%Y-%m-%d %H:%M:%S"


def add_slash(s):
    """
    Adds slash to end of string
    """
    return s if s[-1] == "/" else s + "/"


def remove_slash(s):
    """
    Removes slash to end of string
    """
    return s[0:-1] if s[-1] == "/" else s


def list_to_sql_select(col_list, table_alias=None, exclude=None):
    """
    combines a list of columns to a single string seperated with ', '
    If table alias is provided each column is prefixed with table_name.
    Users can exclude any columns in col_list by adding them in the exclude input
    """
    if table_alias is None:
        table_alias = ""
    else:
        table_alias = table_alias + "."

    if exclude is not None:
        col_list = [c for c in col_list if c not in exclude]

    return ", ".join([table_alias + c for c in col_list])


def read_json(filepath):
    """
    read json from local file
    """
    with open(filepath) as json_data:
        data = json.load(json_data)
    return data


def write_json(data, filepath):
    """
    write json to local file
    """
    with open(filepath, "w") as outfile:
        json.dump(data, outfile, indent=4, separators=(",", ": "))



from pyspark import Row


def apply_cleaning_function_list(pyspark_df, list_of_functions, sampleRatio=None):
    def run_all_functions(row):

        row_dict = row.asDict()

        for fn in list_of_functions:
            row_dict = fn(row_dict)

        return Row(**row_dict)

    new_df = pyspark_df.rdd.map(run_all_functions)
    if sampleRatio:
        return new_df.toDF(sampleRatio=sampleRatio)
    else:
        return new_df.toDF()
